@@ -37,8 +37,9 @@ RCSID("$Id$")
 
 #include <assert.h>
 
-static int retries = 10;
-static float timeout = 3;
+static int success = 0;
+static int retries = 3;
+static float timeout = 5;
 static const char *secret = NULL;
 static int do_output = 1;
 static int totalapp = 0;
@@ -489,6 +490,7 @@ static int send_one_packet(radclient_t *radclient)
 		 *	this packet.
 		 */
 	retry:
+		radclient->request->src_ipaddr.af = AF_UNSPEC;
 		rcode = fr_packet_list_id_alloc(pl, radclient->request);
 		if (rcode < 0) {
 			int mysockfd;
@@ -724,7 +726,12 @@ static int recv_one_packet(int wait_time)
 		       radclient->reply->data_len);
 		vp_printlist(stdout, radclient->reply->vps);
 	}
-	if (radclient->reply->code != PW_AUTHENTICATION_REJECT) {
+
+	if ((radclient->reply->code == PW_AUTHENTICATION_ACK) ||
+	    (radclient->reply->code == PW_ACCOUNTING_RESPONSE) ||
+	    (radclient->reply->code == PW_COA_ACK) ||
+	    (radclient->reply->code == PW_DISCONNECT_ACK)) {
+		success = 1;		/* have a good response */
 		totalapp++;
 	} else {
 		totaldeny++;
@@ -897,6 +904,7 @@ int main(int argc, char **argv)
 	/*
 	 *	Resolve hostname.
 	 */
+	if (force_af == AF_UNSPEC) force_af = AF_INET;
 	server_ipaddr.af = force_af;
 	if (strcmp(argv[1], "-") != 0) {
 		const char *hostname = argv[1];
@@ -960,11 +968,11 @@ int main(int argc, char **argv)
 		packet_code = PW_STATUS_SERVER;
 
 	} else if (strcmp(argv[2], "disconnect") == 0) {
-		if (server_port == 0) server_port = PW_POD_UDP_PORT;
+		if (server_port == 0) server_port = PW_COA_UDP_PORT;
 		packet_code = PW_DISCONNECT_REQUEST;
 
 	} else if (strcmp(argv[2], "coa") == 0) {
-		if (server_port == 0) server_port = PW_POD_UDP_PORT;
+		if (server_port == 0) server_port = PW_COA_UDP_PORT;
 		packet_code = PW_COA_REQUEST;
 
 	} else if (strcmp(argv[2], "auto") == 0) {
@@ -1186,5 +1194,7 @@ int main(int argc, char **argv)
 		printf("\t       Total lost auths:  %d\n", totallost);
 	}
 
-	return 0;
+	if (success) return 0;
+
+	return 1;
 }
