@@ -156,6 +156,10 @@ int eaptls_success(EAP_HANDLER *handler, int peap_flag)
 		if (vps) {
 			SSL_SESSION_set_ex_data(tls_session->ssl->session,
 						eaptls_session_idx, vps);
+		} else {
+			RDEBUG2("WARNING: No information to cache: session caching will be disabled for this session.");
+			SSL_CTX_remove_session(tls_session->ctx,
+					       tls_session->ssl->session);
 		}
 
 		/*
@@ -168,6 +172,7 @@ int eaptls_success(EAP_HANDLER *handler, int peap_flag)
 					     eaptls_session_idx);
 		if (!vp) {
 			RDEBUG("WARNING: No information in cached session!");
+			return eaptls_fail(handler, peap_flag);
 		} else {
 			RDEBUG("Adding cached attributes to the reply:");
 			debug_pair_list(vp);
@@ -176,7 +181,7 @@ int eaptls_success(EAP_HANDLER *handler, int peap_flag)
 			/*
 			 *	Mark the request as resumed.
 			 */
-			vp = pairmake("EAP-Session-Resumed", "0", T_OP_SET);
+			vp = pairmake("EAP-Session-Resumed", "1", T_OP_SET);
 			if (vp) pairadd(&request->packet->vps, vp);
 		}
 	}
@@ -330,7 +335,8 @@ static eaptls_status_t eaptls_ack_handler(EAP_HANDLER *handler)
 		return EAPTLS_FAIL;
 
 	case handshake:
-		if (tls_session->info.handshake_type == finished) {
+		if ((tls_session->info.handshake_type == finished) &&
+		    (tls_session->dirty_out.used == 0)) {
 			RDEBUG2("ACK handshake is finished");
 
 			/* 
@@ -445,7 +451,8 @@ static eaptls_status_t eaptls_verify(EAP_HANDLER *handler)
 		RDEBUG2("Received EAP-TLS ACK message");
 		return eaptls_ack_handler(handler);
 #else
-		if (prev_eap_ds->request->id == eap_ds->response->id) {
+		if (prev_eap_ds &&
+		    (prev_eap_ds->request->id == eap_ds->response->id)) {
 			/*
 			 *	Run the ACK handler directly from here.
 			 */
