@@ -1,7 +1,7 @@
 /*
  * sql_postgresql.c		Postgresql rlm_sql driver
  *
- * Version:	$Id: sql_postgresql.c,v 1.58 2007/12/17 07:45:46 aland Exp $
+ * Version:	$Id$
  *
  *   This program is free software; you can redistribute it and/or modify
  *   it under the terms of the GNU General Public License as published by
@@ -38,7 +38,7 @@
  */
 
 #include <freeradius-devel/ident.h>
-RCSID("$Id: sql_postgresql.c,v 1.58 2007/12/17 07:45:46 aland Exp $")
+RCSID("$Id$")
 
 #include <freeradius-devel/radiusd.h>
 
@@ -270,12 +270,14 @@ static int sql_query(SQLSOCK * sqlsocket, SQL_CONFIG *config, char *querystr) {
 			break;
 
 			case PGRES_FATAL_ERROR:
+#if defined(PG_DIAG_SQLSTATE) && defined(PG_DIAG_MESSAGE_PRIMARY)
 				/*A fatal error occurred.*/
 
 				errorcode = PQresultErrorField(pg_sock->result, PG_DIAG_SQLSTATE);
 				errormsg  = PQresultErrorField(pg_sock->result, PG_DIAG_MESSAGE_PRIMARY);
 				radlog(L_DBG, "rlm_sql_postgresql: Error %s", errormsg);
 				return check_fatal_error(errorcode);
+#endif
 
 			break;
 
@@ -308,6 +310,7 @@ static int sql_query(SQLSOCK * sqlsocket, SQL_CONFIG *config, char *querystr) {
 		}
 		*/
 	}
+	return -1;
 }
 
 
@@ -378,6 +381,25 @@ static int sql_fetch_row(SQLSOCK * sqlsocket, UNUSED SQL_CONFIG *config) {
 	return 0;
 }
 
+/*************************************************************************
+ *
+ *      Function: sql_num_fields
+ *
+ *      Purpose: database specific num_fields. Returns number of rows in
+ *               query
+ *
+ *************************************************************************/
+static int sql_num_fields(SQLSOCK * sqlsocket, UNUSED SQL_CONFIG *config)
+{
+        rlm_sql_postgres_sock *pg_sock = sqlsocket->conn;
+        
+        pg_sock->affected_rows = PQntuples(pg_sock->result);
+        if (pg_sock->result)
+                return PQnfields(pg_sock->result);
+
+        return 0;
+}
+
 
 
 /*************************************************************************
@@ -499,7 +521,7 @@ rlm_sql_module_t rlm_sql_postgresql = {
 	sql_query,
 	sql_select_query,
 	not_implemented, /* sql_store_result */
-	not_implemented, /* sql_num_fields */
+	sql_num_fields,
 	not_implemented, /* sql_num_rows */
 	sql_fetch_row,
 	not_implemented, /* sql_free_result */
